local newList, unpackListAndDel, newDict, unpackDictAndDel
	= Rock:GetRecyclingFunctions("RBM_Config", "newList", "unpackListAndDel", "newDict", "unpackDictAndDel")
local GetNumAddOns = GetNumAddOns
local IsAddOnLoadOnDemand = IsAddOnLoadOnDemand
local IsAddOnLoaded = IsAddOnLoaded
local GetAddOnInfo = GetAddOnInfo
local GetAddOnMetadata = GetAddOnMetadata
local GetRealZoneText = GetRealZoneText
local unpack = unpack
local pairs = pairs
local select = select
local next = next
local LoadAddOn = LoadAddOn

setfenv(1, RBM.FunctionEnviroment)
HasConfigBeenOpen = true
local Core = Core
local MediaHandler = MediaHandler

local function ReturnTrue()
	return true
end

local reservedFrames = {
	_NewBarFrame = true,
	_ForceShow = true,
	_FrameOnClick = true,
	_BarOnClick = true,
	_ColorTables = true,
}

local options = Options

local function GetProfileOpt(key)
	return Profile[key]
end

local function SetProfileOpt(key, value)
	Profile[key] = value
end


options.args.Plugins = {
	type = "group",
	name = L["Plugin options"],
	desc = L["Configure plugin Options."],
	args = {
	},
	order = 10,
	
}

options.args.ShowWelcomeFrame = {
	type = "execute",
	name = L["Show welcome screen"],
	desc = L["Shows the welcome you are intitially greeted with."],
	func = "ShowWelcomeFrame",
	buttonText = L["Show Welcome Screen"],
	order = 5,
}
	
options.args.AllowAnySender = {
	type = "boolean",
	name = L["Allow any sender"],
	desc = L["Allow bars from any sender (within your party or raid), otherwise only raid leaders or officers will be able to send you bars."],
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "AllowAnySender",
	order = 10,
}
		
options.args.TooltipSameZone = {
	type = "boolean",
	name = L["Tooltip same zone"],
	desc = L["Only shows boss modules in the FuBar tooltip that match your current zone."],
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "TooltipSameZone",
	order = 40,
}

options.args.AllowScreenShake = {
	type = "boolean",
	name = L["Allow screen shake"],
	desc = L["Allows the screen to shake briefly to alert you of critical events."],
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "AllowScreenShake",
	order = 50,
}
options.args.ShakeDuration = {
	type = "number",
	name = L["Shake duration"],
	desc = L["How long the screen should shake in seconds."],
	min = .01,
	max = 5,
	step = .01,
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "ShakeDuration",
	order = 51,
}
options.args.ShakeIntensity = {
	type = "number",
	name = L["Shake intensity"],
	desc = L["How intense the screen shake will be."],
	min = 1,
	max = 100,
	step = 1,
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "ShakeIntensity",
	order = 52,
}

options.args.TestShake = {
	type = "execute",
	name = L["Test screen shaking"],
	desc = L["Shakes the screen using the settings provided above."],
	func = "ShakeScreen",
	buttonText = L["Shake screen"],
	order = 53,
}

options.args.AllowFlashFrame = {
	type = "boolean",
	name = L["Allow screen flash"],
	desc = L["Allows the screen to flash briefly to alert you of critical events."],
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "AllowFlashFrame",
	order = 60,
}
options.args.FlashDuration = {
	type = "number",
	name = L["Flash duration"],
	desc = L["How long, in seconds, it should take for the screen should flash."],
	min = .01,
	max = 5,
	step = .01,
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "FlashDuration",
	order = 61,
}

options.args.TestFlash = {
	type = "execute",
	name = L["Test screen flash"],
	desc = L["Flashes the screen using the settings provided above."],
	func = "WarningFlash",
	buttonText = L["Test flash"],
	order = 62,
}

options.args.PlaySoundOnBossDeath = {
	type = "boolean",
	name = L["Play sound on boss death"],
	desc = L["Plays a short sound clip whenever a boss is defeated."],
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "PlaySoundOnBossDeath",
	order = 70,
}
options.args.UseSkull = {
	type = "boolean",
	name = L["Use {Skull} instead of ***"],
	desc = L["Replaces the *** in all announces with Skull Icons"],
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "UseSkull",
	order = 71,
}

do
	local currentAlert
	options.args.AlertColors = {
		type = "group",
		groupType = "inline",
		name = L["Alert colors"],
		desc = L["Sets the color of RBM alerts that show up in the center of the screen."],
		order = 90,
		args = {
			CurrentAlert = {
				type = "choice",
				name = L["Alert"],
				desc = L["Select an alert from the list to change its color."],
				get = function()
					return currentAlert
				end,
				set = function(alert)
					currentAlert = alert
				end,
				order = 10,
				choices = function()
					local dict = newDict()
					for alert in pairs(Profile.AlertColors) do
						dict[alert] = L[alert]
					end
					return "@dict", unpackDictAndDel(dict)
				end,
			},
			Color = {
				type = "color",
				name = L["Alert color"],
				desc = L["Sets the color of the alert selected above."],
				disabled = function()
					return not currentAlert
				end,
				set = function(r, g, b)
					if currentAlert then
						local a = Profile.AlertColors[currentAlert]
						a.r, a.g, a.b = r, g, b
					end
				end,
				get = function()
					if currentAlert then
						local a = Profile.AlertColors[currentAlert]
						return a.r, a.g, a.b
					end
				end,
				order = 20,
			},
		},
	}
end

options.args.ShowAlerts = {
	type = "boolean",
	name = L["Show alerts"],
	desc = L["Displays messages in the center of the screen for important events."],
	get = GetProfileOpt,
	set = SetProfileOpt,
	passValue = "ShowAlerts",
	order = 80,
}

options.args.BossModules = {
	type = "group",
	name = L["BossModule options"],
	desc = L["Configure boss module Options."],
	args = {
		_LoDModules = {
			type = "multichoice",
			name = L["LoD Modules"],
			desc = L["Lists all unloaded LoD modules, and optionally allows you to load them early."],
			set = LoadAddOn,
			get = Noop,
			choices = function()
				local list = newList()
				for i=1, GetNumAddOns() do
					if IsAddOnLoadOnDemand(i) and not IsAddOnLoaded(i) then
						local enabled, loadable = select(4, GetAddOnInfo(i))
						local zones = GetAddOnMetadata(i, "X-RBMLoadIn")
						if zones and enabled and loadable then
							list[#list + 1] = GetAddOnInfo(i)
						end
					end
				end
				return "@list", unpackListAndDel(list)
			end,
			order = 10,
		},
		_ActivateModulesZone = {
			type = "boolean",
			name = L["Activate modules on zone match"],
			desc = L["Automatically enables boss modules when you enter a zone they belong in."],
			get = GetProfileOpt,
			set = SetProfileOpt,
			passValue = "ActivateModulesZone",
			order = 20,
		},
		_ActivateModulesMouseover = {
			type = "boolean",
			name = L["Activate modules on mouseover"],
			desc = L["Automatically enables boss modules with the same name as your mouseover target."],
			get = GetProfileOpt,
			set = SetProfileOpt,
			passValue = "ActivateModulesMouseover",
			order = 30,
		},
	},
	order = 10,
}

local rightClickChoices = {
	TestBar = L["Show a test bar"],
	ShowMenu = L["Open the configuration menu"],
	CancelAll = L["Cancel all bars"],
	Nothing = L["Do nothing"],
}

local barClickChoices = {
	Cancel = L["Cancels the bar"],
	ShowMenu = L["Opens the configuration menu to bar options"],
	Pause = L["Pause/unpause bar"],
	Filter = L["Filter bar"],
	Color = L["Recolor bar"],
	Nothing = L["Do nothing"],
}

do
	local currentCategory, currentSubcategory
	options.args.Frames = {
		type = "group",
		name = L["Frame options"],
		desc = L["Configure frame options."],
		order = 20,
		args = {
			_NewBarFrame = {
				type = "string",
				name = L["New frame"],
				desc = L["Creates a new frame with the specified name."],
				get = false,
				set = function(name)
					Core:NewBarFrame(name, true)
					Core:RefreshConfig()
					Core:OpenConfigMenu(L["Frames"], name)
				end,
				order = 10,
				validate = function(name)
					if name == "" then
						return false, L["The frame name cannot be empty."]
					elseif Core:HasBarFrame(name) then
						return false, L["A frame with that name already exists."]
					elseif reservedFrames[name] then
						return false, L["This name is reserved."]
					end
					return true
				end,
				usage = L["Enter the name of the frame you wish to create and press enter."],
			},
			_ForceShow = {
				type = "keybinding",
				name = L["Force show"],
				desc = L["Sets a keybinding that forces all frames to be shown regardless of their settings."],
				set = function(bind)
					local oldBind = _G.GetBindingKey("RBM_FORCE_SHOW")
					if oldBind then
						_G.SetBinding(oldBind, nil)
					end
					_G.SetBinding(bind, "RBM_FORCE_SHOW")
					_G.SaveBindings(_G.GetCurrentBindingSet())
				end,
				get = function() return _G.GetBindingKey("RBM_FORCE_SHOW") end,
				order = 20,
			},
			_FrameOnClick = {
				type = "group",
				name = L["Frame right-click events"],
				desc = L["Determines what happens when frames are right-clicked on."],
				groupType = "inline",
				child_get = GetProfileOpt,
				child_set = SetProfileOpt,
				child_choices = rightClickChoices,
				order = 30,
				args = {
					OnRightClick = {
						type = "choice",
						name = L["Right-click"],
						desc = L["Determines what happens when a frame is right-clicked on."],
						passValue = "FrameRightClick",
						order = 10,
					},
					OnControlRightClick = {
						type = "choice",
						name = L["Control + Right-click"],
						desc = L["What happens when a frame is Control + Right-clicked."],
						passValue = "FrameControlRightClick",
						order = 20,
					},
					OnShiftRightClick = {
						type = "choice",
						name = L["Shift + Right-click"],
						desc = L["What happens when a frame is Shift + Right-clicked."],
						passValue = "FrameShiftRightClick",
						order = 30,
					},
				},
			},
			_BarOnClick = {
				type = "group",
				name = L["Bar click events"],
				desc = L["Determines what happens when a bar is clicked on."],
				groupType = "inline",
				child_get = GetProfileOpt,
				child_set = SetProfileOpt,
				child_choices = barClickChoices,
				order = 40,
				args = {
					BarLeftClick = {
						type = "choice",
						name = L["Left-click"],
						desc = L["What happens when a bar is left-clicked."],
						passValue = "BarLeftClick",
						order = 10,
					},
					BarRightClick = {
						type = "choice",
						name = L["Right-click"],
						desc = L["What happens when a bar is right-clicked."],
						passValue = "BarRightClick",
						order = 20,
					},
					BarMiddleClick = {
						type = "choice",
						name = L["Middle-click"],
						desc = L["What happens when a frame is shift right-clicked."],
						passValue = "BarMiddleClick",
						order = 30,
					},
				},
			},
			_ColorTables = {
				type = "group",
				name = L["Bar colors"],
				desc = L["All bars obey these colors (unless they've been custom colored)."],
				order = 50,
				groupType = "inline",
				args = {
					CurrentCategory = {
						type = "choice",
						name = L["Category"],
						desc = L["Select one of the categories from the dropdown to recolor it."],
						get = function()
							if not currentCategory then
								currentCategory = "Boss Abilities"--#NO_LOCALIZE
								currentSubcategory = "Cast"--#NO_LOCALIZE
							end
							return currentCategory
						end,
						set = function(category)
							currentCategory = category
							if currentCategory == "Boss Abilities" then--#NO_LOCALIZE
								currentSubcategory = next(Profile.ColorTables[currentCategory])
							else
								currentSubcategory = next(Core:GetModule("Plugin", currentCategory):GetCategories())
							end
						end,
						order = 10,
						choices = function()
							local dict = newDict()
							dict["Boss Abilities"]--#NO_LOCALIZE
								= L["Boss Abilities"]
							for name, module in Core:IterateModules("Plugin") do
								if module:GetCategories() then
									dict[name] = name
								end
							end
							return "@dict", unpackDictAndDel(dict)
						end,
					},
					CurrentSubCategory = {
						type = "choice",
						name = L["Sub-category"],
						desc = L["Select one of the sub-categories from the dropdown to recolor it."],
						get = function()
							return currentSubcategory
						end,
						set = function(subCategory)
							currentSubcategory = subCategory
						end,
						order = 20,
						choices = function()
							local dict = newDict()
							if currentCategory == "Boss Abilities" then--#NO_LOCALIZE
								for subCat in pairs(Profile.ColorTables[currentCategory]) do
									dict[subCat] = L[subCat]
								end
							else
								local module = Core:GetModule("Plugin", currentCategory)
								for subCat in pairs(module:GetCategories()) do
									dict[subCat] = subCat
								end
							end
							return "@dict", unpackDictAndDel(dict)
						end,
					},
					Color = {
						type = "color",
						name = L["Color"],
						desc = L["Sets the color of bars that belong to the above selections."],
						set = function(r, g, b)
							local ct
							if currentCategory == "Boss Abilities" then--#NO_LOCALIZE
								ct = Profile.ColorTables[currentCategory][currentSubcategory]
								
							else
								ct = Profile.ColorTables.Modules[currentCategory][currentSubcategory]
							end
							ct[1], ct[2], ct[3] = r, g, b
						end,
						get = function()
							if currentCategory and currentSubcategory then
								if currentCategory == "Boss Abilities" then--#NO_LOCALIZE
									return unpack(Profile.ColorTables[currentCategory][currentSubcategory])
									
								else
									return unpack(Profile.ColorTables.Modules[currentCategory][currentSubcategory])
								end
							end
						end,
						order = 30,
					},
				},
			},
		},
	}
end

local FrameOptionsTable



local function AddToggles(module, toggles)
	local opts = module:GetOptionsTable().args

	for toggle in pairs(toggles) do
		if not opts[toggle] then
			opts[toggle] = {
				handler = module,
				type = "boolean",
				name = toggle,
				desc = toggle,
				get = "CheckToggle",
				set = "SetToggle",
				hidden = "GetHideQuickCommands",
				passValue = toggle,
				order = 1000,
			}
		end
		
	end
end

local function AddExecutes(module, executes)
	local opts = module:GetOptionsTable().args
	for execute, func in pairs(executes) do
		if not opts[execute] then
			opts[execute] = {
				handler = module,
				type = "execute",
				name = execute,
				desc = execute,
				func = func,
				hidden = "GetHideQuickCommands",
				passValue = module,
				order = 2000,
			}
			if execute == L["Force Start"] then
				opts[execute].disabled = "~CanStart"
			elseif execute == L["Force Stop"] then
				opts[execute].disabled = "~CanStop"
			end
		end
		
	end
end



local ToggleActiveArgs 
local LoDModules = LoDModules

if LoDModules then
	Core:AddEventListener("ADDON_LOADED", function(_, _, addon)
		local plugin = GetAddOnMetadata(addon, "X-RBMPlugin")
		if plugin then
			LoDModules[plugin] = nil
			options.args.Plugins.args[plugin] = nil
			local module = Core:GetModule("Plugin", plugin)
			Profile.ModuleStates[module:GetPrefixedName()] = true
			if not next(LoDModules) then
				LoDModules = nil
				ToggleActiveArgs = nil
				Core:RemoveEventListener("ADDON_LOADED")
			end
			Core:RefreshConfig()
			Core:AddTimer(0, "OpenConfigMenu", "Plugins", plugin)
		end
	end)
end
local AnnounceAll, DisableAll, EnableAll
function Core:RefreshConfig()
	local plugins = options.args.Plugins.args
	for name, module in self:IterateModules("Plugin") do
		if LoDModules and LoDModules[name] then
			LoDModules[name] = nil
			plugins[name] = nil
		end
		if not plugins[name] then
			local toggles = module:GetToggles()
			if toggles then
				AddToggles(module, toggles)
			end
			local executes = module:GetExecutes()
			if executes then
				AddExecutes(module, executes)
			end
			plugins[name] = module:GetOptionsTable(true)
		end
	end
	if LoDModules then
		for name, addon in pairs(LoDModules) do
			if not plugins[name] then
				if not ToggleActiveArgs then
					ToggleActiveArgs = {
						args = {
							type = "boolean",
							name = L["Enabled"],
							desc = L["This module is Load on Demand and currently unloaded because it's disabled, enable this module to automatically load it."],
							set = LoadAddOn,
							get = false,
						}
					}
				end
				plugins[name] = {
					type = "group",
					name = name,
					desc = name,
					child_passValue = addon,
					args = ToggleActiveArgs,
				}
			end
		end
	end
	local zones = options.args.BossModules.args
	for name, module in self:IterateModules("BossModule") do
		local zone = module:GetDisplayZone()
		if not zones[zone] then
			if not AnnounceAll then
				AnnounceAll = {
					type = "boolean",
					name = L["Announce all"],
					desc = L["Sets all modules that belong to this zone to announce to the raid."],
					set = function(zone)
						local announce = true
						for name, module in self:IterateModules("BossModule") do
							if module:HasZone(zone) then
								if not module:CheckToggle(L["Announce"]) then
									announce = false
									break
								end
							end
						end
						for name, module in self:IterateModules("BossModule") do
							if module:HasZone(zone) then
								module:SetToggle(L["Announce"], not announce)
							end
						end
					end,
					get = function(zone)
						local announce = false
						for name, module in self:IterateModules("BossModule") do
							if module:HasZone(zone) then
								if not module:CheckToggle(L["Announce"]) then
									if announce then
										return "HALF"
									end
								else
									announce = true
								end
							end
						end
						return announce
					end,
					order = 10,
				}
				EnableAll = {
					type = "execute",
					name = L["Enable all"],
					desc = L["Enables all modules that belong to this zone."],
					func = function(zone)
						for name, module in self:IterateModules("BossModule") do
							if module:HasZone(zone) then
								module:Enable()
							end
						end
					end,
					buttonText = L["Enable all"],
					order = 20,
				}
				DisableAll = {
					type = "execute",
					name = L["Disable all"],
					desc = L["Disable all modules that belong to this zone."],
					func = function(zone)
						for name, module in self:IterateModules("BossModule") do
							if module:HasZone(zone) then
								module:Disable()
							end
						end
					end,
					buttonText = L["Disable all"],
					order = 30,
				}
			end
			zones[zone] = {
				type = "group",
				name = zone,
				desc = L["Options for modules in this zone."],
				child_passValue = zone,
				args = {
					_AnnounceAll = AnnounceAll,
					_EnableAll = EnableAll,
					_DisableAll = DisableAll,
				},
			}
		end
		local args = zones[zone].args
		if not args[name] then
			local toggles = module:GetToggles()
			if toggles then
				AddToggles(module, toggles)
			end
			local executes = module:GetExecutes()
			if executes then
				AddExecutes(module, executes)
			end
			args[name] = module:GetOptionsTable(true)
		end
	end
	local frames = options.args.Frames.args
	for name in pairs(frames) do
		if not reservedFrames[name] and not self:HasBarFrame(name) then --remove deleted frames
			frames[name] = nil
		end
	end
	for name, frame in self:IterateBarFrames() do
		if not frames[name] then
			frames[name] = {
				type = "group",
				desc = L["Configures this frame."],
				name = name,
				handler = frame,
				args = FrameOptionsTable,
			}
		end
	end
	_G.Rock("LibRockConfig-1.0"):RefreshConfigMenu(Core)
end

local blendModes = {
	DISABLE = L["Opaque texture"],
	BLEND = L["Normal painting on top of the background with alpha"],
	ALPHAKEY = L["One-bit alpha"],
	ADD = L["Additive blend (Default)"],
	MOD = L["Modulating blend"],
}

local transferAnimations = {
	None			=	L["No animation"],
	CurveDownConstant	=	L["Curve down, constant speed"],
	CurveUpConstant		=	L["Curve up, constant speed"],
	CurveUpPower		=	L["Curve up, starts fast then slows"],
	CurveDownPower		=	L["Curve down, starts fast then slows"],
	XThenY			=	L["Right-angle, finishing X movement first"],
	YThenX			=	L["Right-angle, finishing Y movement first"],
	SlidePower		=	L["Slide, starts fast then slows"],
	SlideConstant		=	L["Slide, constant speed"],
}

local themeOpts = {
	Sound = true,
	Filters = true,
	Transfer = false,
	Elements = true,
}
local function SetThemeOption(option, value)
	themeOpts[option] = value
end
local function GetThemeOption(option)
	return themeOpts[option]
end

local moduleCategories = {}

local function RefreshConfigModuleCategories()
	for name, module in Core:IterateModules("Plugin") do
		if not moduleCategories[name] then
			local cats = module:GetCategories()
			if cats then
				local t = {
					type = "multichoice",
					name = name,
					choices = {},
					choiceDescs = {},
					passValue3 = name,
				}
				for category in pairs(cats) do
					t.choices[category] = category
					t.choiceDescs[category] = module:GetCategoryDescription(category)
				end
				moduleCategories[name] = t
			end
		else
			local choices = moduleCategories[name].choices
			local choiceDescs = moduleCategories[name].choiceDescs
			for category in pairs(module:GetCategories()) do
				if not choices[category] then
					choices[category] = category
					choiceDescs[category] = module:GetCategoryDescription(category)
				end
			end
		end
	end
	return moduleCategories
end

FrameOptionsTable = {
	Destroy = {
		type = "execute",
		name = L["Destroy frame"],
		desc = L["Destroys a frame, removing all of its settings permanently."],
		func = "GUIDestroy",
		confirmText = L["Are you sure?"],
		buttonText = L["Destroy frame"],
		order = 10,
	},
	Reset = {
		type = "execute",
		name = L["Reset frame"],
		desc = L["Resets a frame, setting all of its options back to default."],
		func = "Reset",
		confirmText = L["Are you sure?"],
		buttonText = L["Reset frame"],
		order = 15,
	},
	ResetAnchor = {
		type = "execute",
		name = L["Reset frame anchor"],
		desc = L["Resets just the frame's anchor, placing it in the center of the screen."],
		func = "ResetAnchor",
		confirmText = L["Are you sure?"],
		buttonText = L["Reset frame anchor"],
		order = 17,
	},
	Basics = {
		type = "group",
		name = L["Basics"],
		desc = L["Basic options that control the behavior of the frame."],
		groupType = "inline",
		order = 20,
		args = {
			Hide = {
				type = "boolean",
				name = L["Show frame background"],
				desc = L["If not checked the frame will be hidden, but the bars will still display."],
				set = "ToggleVisible",
				get = "IsVisible",
				order = 20,
			},
			Lock = {
				type = "boolean",
				name = L["Lock frame"],
				desc = L["The frame will not be draggable."],
				set = "ToggleLock",
				get = "IsLocked",
				order = 30,
			},
			Title = {
				type = "boolean",
				name = L["Show title"],
				desc = L["The title of the frame will be shown."],
				set = "SetOption",
				get = "GetOption",
				order = 40,
				passValue = "ShowTitle",
			},
			DynamicResize = {
				type = "boolean",
				name = L["Dynamic resize"],
				desc = L["The frame will automatically adjust it's size to fit the bars on the frame, up to the maximum size defined by dragging the resize handle."],
				set = "SetOption",
				get = "GetOption",
				order = 50,
				passValue = "DynamicResize",
			},
			HideWhenEmpty = {
				type = "boolean",
				name = L["Hide when empty"],
				desc = L["Hides the frame when it's empty, you can force it to be displayed by setting a keybind in 'Frame Options' or by unchecking this."],
				set = "SetOption",
				get = "GetOption",
				order = 60,
				passValue = "HideWhenEmpty",
			},
			Alpha = {
				type = "number",
				name = L["Frame transparency"],
				desc = L["Sets the overall transparency of the frame, including all bars and text belonging to the frame."],
				set = "SetOption",
				get = "GetOption",
				step = .01,
				bigStep = .05,
				isPercent = true,
				order = 70,
				passValue = "Alpha",
			},
		},
	},
	Coloring = {
		type = "group",
		name = L["Coloring"],
		desc = L["Options that control the color of the frame."],
		groupType = "inline",
		order = 30,
		child_get = "GetOption",
		child_set = "SetOption",
		args = {
			BackdropColor = {
				type = "color",
				name = L["Backdrop"],
				desc = L["Sets the color of the backdrop for this frame."],
				hasAlpha = true,
				passValue = "colorTable",
				passValue2 = "BackdropColor",
				order = 10,
			},
			BackdropBorderColor = {
				type = "color",
				name = L["Border"],
				desc = L["Sets the color of the backdrop border for this frame."],
				hasAlpha = true,
				passValue = "colorTable",
				passValue2 = "BackdropBorderColor",
				order = 20,
			},
		},
	},
	Textures = {
		type = "group",
		name = L["Textures"],
		desc = L["Options that control the textures of the frame."],
		groupType = "inline",
		order = 40,
		child_get = "GetOption",
		child_set = "SetOption",
		child_passValue = "Backdrop",
		args = {
			bgFile = {
				type = "choice",
				name = L["Background texture"],
				desc = L["Sets the background texture for this frame to use."],
				choices = MediaHandler:List("background"),
				passValue2 = "bgFile",
				order = 10,
			},
			edgeFile = {
				type = "choice",
				name = L["Border texture"],
				desc = L["Sets the border texture for this frame to use."],
				choices = MediaHandler:List("border"),
				passValue2 = "edgeFile",
				order = 20,
			},
			edgeSize = {
				type = "number",
				name = L["Border size"],
				desc = L["Sets the size of the border texture for this frame."],
				step = 1,
				bigStep = 5,
				stepBasis = 5,
				min = 1,
				max = 50,
				passValue2 = "edgeSize",
				order = 30,
			},
			tile = {
				type = "boolean",
				name = L["Tile background"],
				desc = L["Whether or not the background should tile instead of stretching."],
				passValue2 = "tile",
				order = 40,
			},
			Insets = {
				type = "group",
				name = L["Insets"],
				desc = L["Stretches or shrinks the background texture."],
				groupType = "inline",
				child_passValue = "Backdrop",
				child_passValue2 = "insets",
				child_get = "GetOption",
				child_set = "SetOption",
				order = 50,
				args = {
					left = {
						type = "number",
						name = L["Left"],
						desc = L["Stretches or shrinks the background texture."],
						step = 1,
						bigStep = 5,
						min = -50,
						max = 50,
						passValue3 = "left",
						stepBasis = 5,
						order = 10,
					},
					right = {
						type = "number",
						name = L["Right"],
						desc = L["Stretches or shrinks the background texture."],
						step = 1,
						bigStep = 5,
						min = -50,
						max = 50,
						passValue3 = "right",
						stepBasis = 5,
						order = 20,
					},
					top = {
						type = "number",
						name = L["Top"],
						desc = L["Stretches or shrinks the background texture."],
						step = 1,
						bigStep = 5,
						min = -50,
						max = 50,
						passValue3 = "top",
						stepBasis = 5,
						order = 30,
					},
					bottom = {
						type = "number",
						name = L["Bottom"],
						desc = L["Stretches or shrinks the background texture."],
						step = 1,
						bigStep = 5,
						min = -50,
						max = 50,
						passValue3 = "bottom",
						stepBasis = 5,
						order = 40,
					},
				},
			},
			
		},
	},
	BarControl = {
		type = "group",
		desc = L["Control options such as canceling, testing, sorting and transferring bars."],
		name = L["Bar control"],
		order = 50,
		child_set = "SetOption",
		child_get = "GetOption",
		args = {
			Basics = {
				type = "group",
				name = L["Basics"],
				desc = L["Basic options such as canceling, testing, and clicking bars."],
				groupType = "inline",
				order = 10,
				child_get = "GetOption",
				child_set = "SetOption",
				args = {
					CancelAll = {
						type = "execute",
						name = L["Cancel all bars"],
						desc = L["Cancels all bars currently running on this frame."],
						func = "CancelAllBars",
						buttonText = L["Cancel all"],
						order = 10,
					},
					DisplayTest = {
						type = "execute",
						name = L["Display a test bar"],
						desc = L["Creates and displays a test bar on this frame."],
						func = "TestBar",
						buttonText = L["Display a bar"],
						order = 20,
					},
					DisplayMaxTest = {
						type = "execute",
						name = L["Display max test bars"],
						desc = L["Creates and displays multiple test bars."],
						func = "MaxTestBars",
						buttonText = L["Display a bunch!"],
						order = 30,
					},
					TestIgnoresFilter = {
						type = "boolean",
						name = L["Test bars should ignore filters"],
						desc = L["When testing bars they're still appear even if they don't pass filters."],
						passValue = "TestIgnoresFilter",
						order = 60,
					},
					EnableMouse = {
						type = "boolean",
						name = L["Clickable bars"],
						desc = L["If checked then bars will be clickable, otherwise mouse clicks will pass through them."],
						passValue = "Bar",
						passValue2 = "EnableMouse",
						order = 80,
					},
				},
			},
			BarSorting = {
				type = "group",
				name = L["Sorting"],
				desc = L["Configures bar sorting options."],
				order = 90,
				groupType = "inline",
				child_set = "SetOption",
				child_get = "GetOption",
				args = {
					Sorting = {
						type = "boolean",
						name = L["Enabled"],
						desc = L["If checked then bar sorting will be active."],
						order = 10,
						passValue = "SortBars",
					},
					SortOrder = {
						type = "choice",
						name = L["Order"],
						desc = L["The order in which bars are sorted."],
						choices = {
							Up = L["Up"], 
							Down = L["Down"],
						},
						disabled = "~IsSortingEnabled",
						passValue = "SortBarsOrder",
						order = 20,
					},
					SortMethod = {
						type = "choice",
						name = L["Method"],
						desc = L["The method in which the bars are sorted by."],
						choices = {
							["Time Left"]--#NO_LOCALIZE
								= L["Time: Current Left"],
							["Time Total Length"]--#NO_LOCALIZE
								= L["Time: Total Length"],
							["Label"]--#NO_LOCALIZE
								= L["Label: Text"],
							["Category"]--#NO_LOCALIZE
								= L["Category"],
							["SubCategory"]	--#NO_LOCALIZE
								= L["SubCategory"]
						},
						disabled = "~IsSortingEnabled",
						passValue = "SortBarsMethod",
						order = 30,
					},
				},
			},
			BarTransferring = {
				type = "group",
				name = L["Transferring"],
				desc = L["Configures bar transferring options."],
				order = 100,
				groupType = "inline",
				args = {
					Above = {
						type = "group",
						name = L["Above a threshold"],
						desc = L["Configures bar transferring options when a bar goes above a certain amount of time."],
						order = 10,
						groupType = "inline",
						child_set = "SetOption",
						child_get = "GetOption",
						args = {
							TransferTo = {
								type = "choice",
								name = L["Transfer to"],
								desc = L["Sets the frame to which the bars will be transferred if the meet the criteria described below."],
								choices = "GetTransferFrames",
								passValue = "TransferAboveTo",
								order = 10,
							},
							TransferByPercent = {
								type = "boolean",
								name = L["Transfer by percent"],
								desc = L["Enables bar transfers when bars are above a certain percentage in time. For example, if this is set to 50% and a 2 minute bar goes above 1 minute left the bar will be transferred."],
								order = 20,
								passValue = "TransferAboveByPercent",
							},
							TransferPercent = {
								type = "number",
								name = L["Transfer percentage"],
								desc = L["If the bar time goes above this percentage of time it'll be transferred."],
								isPercent = true,
								step = .01,
								bigStep = .05,
								stepBasis = .05,
								passValue = "TransferAbovePercent",
								order = 30,
							},
							TransferByTime = {
								type = "boolean",
								name = L["Transfer by absolute time"],
								desc = L["Enables bar transfers when bars are above a certain amount of seconds. If transfer by percent is also enabled whichever criteria is met first will trigger the transfer."],
								order = 40,
								passValue = "TransferAboveByTime",
							},
							TransferTime = {
								type = "number",
								name = L["Transfer time"],
								desc = L["If the bar's time left goes above this time in seconds it'll be transferred."],
								step = 1,
								bigStep = 5,
								stepBasis = 5,
								min = 0,
								max = 600,
								passValue = "TransferAboveTime",
								order = 50,
							},
							TransferAnimate = {
								type = "choice",
								name = L["Transfer animation"],
								desc = L["Selects the animation used when transferring bars."],
								order = 60,
								choices = transferAnimations,
								passValue = "TransferAboveAnimate",
							},
							TransferRate = {
								type = "number",
								name = L["Animation speed"],
								desc = L["How quickly the bar will transfer to the other frame if an animation is selected. Speed effects each animation differently: 100% is always very fast, or instant, and 1% is very slow."],
								step = .01,
								min = .01,
								isPercent = true,
								passValue = "TransferAboveRate",
								order = 70,
							},
						},
					},
					Below = {
						type = "group",
						name = L["Below a threshold"],
						desc = L["Configures bar transferring options when a bar goes below a certain amount of time."],
						order = 20,
						groupType = "inline",
						child_set = "SetOption",
						child_get = "GetOption",
						args = {
							TransferTo = {
								type = "choice",
								name = L["Transfer to"],
								desc = L["Sets the frame to which the bars will be transferred if the meet the criteria described below."],
								choices = "GetTransferFrames",
								passValue = "TransferBelowTo",
								order = 10,
							},
							TransferByPercent = {
								type = "boolean",
								name = L["Transfer by percent"],
								desc = L["Enables bar transfers when bars are below a certain percentage in time. For example, if this is set to 50% and a 2 minute bar goes below 1 minute left the bar will be transferred."],
								order = 20,
								passValue = "TransferBelowByPercent",
							},
							TransferPercent = {
								type = "number",
								name = L["Transfer percentage"],
								desc = L["If the bar time goes below this percentage of time it'll be transferred."],
								isPercent = true,
								step = .01,
								bigStep = .05,
								stepBasis = .05,
								passValue = "TransferBelowPercent",
								order = 30,
							},
							TransferByTime = {
								type = "boolean",
								name = L["Transfer by absolute time"],
								desc = L["Enables bar transfers when bars are below a certain amount of seconds. If transfer by percent is also enabled whichever criteria is met first will trigger the transfer."],
								order = 40,
								passValue = "TransferBelowByTime",
							},
							TransferTime = {
								type = "number",
								name = L["Transfer time"],
								desc = L["If the bar's time left goes below this time in seconds it'll be transferred."],
								step = 1,
								bigStep = 5,
								stepBasis = 5,
								min = 0,
								max = 600,
								passValue = "TransferBelowTime",
								order = 50,
							},
							TransferAnimate = {
								type = "choice",
								name = L["Transfer animation"],
								desc = L["Selects the animation used when transferring bars."],
								order = 60,
								choices = transferAnimations,
								passValue = "TransferBelowAnimate",
							},
							TransferRate = {
								type = "number",
								name = L["Animation speed"],
								desc = L["How quickly the bar will transfer to the other frame if an animation is selected. Speed effects each animation differently: 100% is always very fast, or instant, and 1% is very slow."],
								step = .01,
								min = .01,
								isPercent = true,
								passValue = "TransferBelowRate",
								order = 70,
							},
						},
					},
				},
			},
			BarBlinking = {
				type = "group",
				name = L["Blinking"],
				desc = L["Allows the bars to blink in emphasis under certain conditions."],
				order = 110,
				groupType = "inline",
				child_set = "SetOption",
				child_get = "GetOption",
				args = {
					BlinkByPercent = {
						type = "boolean",
						name = L["Blink by percent"],
						desc = L["Enables bar blinking when bars are under a certain percentage in time. For example, if this is set to 50% and a 2 minute bar goes under 1 minute left the bar will start blinking."],
						order = 10,
						passValue = "BlinkByPercent",
					},
					BlinkPercent = {
						type = "number",
						name = L["Blink percentage"],
						desc = L["If the bar time goes under this percentage of time it'll start blinking."],
						isPercent = true,
						step = .01,
						bigStep = .05,
						stepBasis = .05,
						passValue = "BlinkPercent",
						order = 20,
					},
					BlinkByTime = {
						type = "boolean",
						name = L["Blink by absolute time"],
						desc = L["Enables bar blinking when bars are under a certain amount of seconds. If blink by percent is also enabled whichever criteria is met first will trigger the blinking."],
						order = 30,
						passValue = "BlinkByTime",
					},
					BlinkTime = {
						type = "number",
						name = L["Blink time"],
						desc = L["If the bar time goes under this time in seconds it'll start blinking."],
						step = 1,
						bigStep = 5,
						stepBasis = 5,
						min = 0,
						max = 600,
						passValue = "BlinkTime",
						order = 40,
					},
				},
			},
			BarPulsing = {
				type = "group",
				name = L["Pulsing"],
				desc = L["Allows the bars to pulse in emphasis under certain conditions."],
				order = 120,
				groupType = "inline",
				child_set = "SetOption",
				child_get = "GetOption",
				args = {
					PulseByPercent = {
						type = "boolean",
						name = L["Pulse by percent"],
						desc = L["Enables bar pulsing when bars are under a certain percentage in time. For example, if this is set to 50% and a 2 minute bar goes under 1 minute left the bar will start pulsing."],
						order = 10,
						passValue = "PulseByPercent",
					},
					PulsePercent = {
						type = "number",
						name = L["Pulse percentage"],
						desc = L["If the bar time goes under this percentage of time it'll start pulsing."],
						isPercent = true,
						step = .01,
						bigStep = .05,
						stepBasis = .05,
						passValue = "PulsePercent",
						order = 20,
					},
					PulseByTime = {
						type = "boolean",
						name = L["Pulse by absolute time"],
						desc = L["Enables bar pulsing when bars are under a certain amount of seconds. If pulse by percent is also enabled whichever criteria is met first will trigger the pulsing."],
						order = 30,
						passValue = "PulseByTime",
					},
					PulseTime = {
						type = "number",
						name = L["Pulse time"],
						desc = L["If the bar time goes under this time in seconds it'll start pulsing."],
						step = 1,
						bigStep = 5,
						stepBasis = 5,
						min = 0,
						max = 600,
						passValue = "PulseTime",
						order = 540,
					},
				},
			},
		},
	},
	BarOptions = {
		type = "group",
		desc = L["Bar visual options like coloring, textures, anchors, and visibility."],
		name = L["Bar options"],
		order = 60,
		args = {
			Basics = {
				type = "group",
				name = L["Basics"],
				desc = L["Basic options such as sizing, anchoring and spacing."],
				groupType = "inline",
				order = 20,
				child_get = "GetOption",
				child_set = "SetOption",
				child_passValue = "Bar",
				args = {
					FrameAnchor = {
						type = "choice",
						name = L["Bar growth"],
						desc = L["Controls which direction bars grow from."],
						order = 10,
						passValue2 = "FrameAnchor",
						choices = {
							BOTTOM = L["From the bottom"],
							TOP = L["From the top"],
						},
					},
					Width = {
						type = "number",
						name = L["Width"],
						desc = L["Sets the width of a bar relative to the width of the frame it belongs to. 100% means the bar would be the entire width of the frame, 50% means the bar would be half the width of the frame."],
						step = .01,
						bigStep = .05,
						stepBasis = .05,
						min = .3,
						isPercent = true,
						passValue2 = "Width",
						order = 20,
					},
					Height = {
						type = "number",
						name = L["Height"],
						desc = L["Sets the absolute height of a bar. The bar cannot be larger than the half the height of the frame."],
						step = 1,
						bigStep = 5,
						stepBasis = 5,
						min = 5,
						max = "GetMaxBarHeight",
						passValue2 = "Height",
						order = 30,
					},
					Spacing = {
						type = "number",
						name = L["Spacing"],
						desc = L["Sets the absolute amount of space between bars."],
						step = 1,
						bigStep = 2,
						stepBasis = 2,
						min = -10,
						max = 10,
						passValue2 = "Spacing",
						order = 40,
					},
				},
			},
			Coloring = {
				type = "group",
				name = L["Coloring"],
				desc = L["Options that control the color of the bars belonging to this frame."],
				groupType = "inline",
				order = 30,
				child_get = "GetOption",
				child_set = "SetOption",
				child_passValue = "colorTable",
				child_passValue2 = "Bar",
				args = {
					BackdropColor = {
						type = "color",
						name = L["Backdrop"],
						desc = L["Sets the color of the backdrop for bars belonging to this frame."],
						hasAlpha = true,
						
						passValue3 = "BackdropColor",
						order = 10,
					},
					BackdropBorderColor = {
						type = "color",
						name = L["Border"],
						desc = L["Sets the color of the backdrop border for bars belonging to this frame."],
						hasAlpha = true,
						passValue3 = "BackdropBorderColor",
						order = 20,
					},
				},
			},
			Textures = {
				type = "group",
				name = L["Textures"],
				desc = L["Options that control the textures of the bars belonging to this frame."],
				groupType = "inline",
				order = 40,
				child_get = "GetOption",
				child_set = "SetOption",
				child_passValue = "Bar",
				child_passValue2 = "Backdrop",
				args = {
					bgFile = {
						type = "choice",
						name = L["Background texture"],
						desc = L["Sets the background texture for bars to use."],
						choices = MediaHandler:List("background"),
						passValue3 = "bgFile",
						order = 10,
					},
					edgeFile = {
						type = "choice",
						name = L["Border texture"],
						desc = L["Sets the border texture for bars to use."],
						choices = MediaHandler:List("border"),
						passValue3 = "edgeFile",
						order = 20,
					},
					edgeSize = {
						type = "number",
						name = L["Border size"],
						desc = L["Sets the size of the border texture for bars belonging to this frame."],
						step = 1,
						bigStep = 5,
						stepBasis = 5,
						min = 1,
						max = 50,
						passValue3 = "edgeSize",
						order = 30,
					},
					tile = {
						type = "boolean",
						name = L["Tile background"],
						desc = L["Whether or not the background should tile instead of stretching."],
						passValue3 = "tile",
						order = 40,
					},
					Insets = {
						type = "group",
						name = L["Insets"],
						desc = L["Stretches or shrinks the background texture."],
						groupType = "inline",
						child_passValue = "Bar",
						child_passValue2 = "Backdrop",
						child_passValue3 = "insets",
						child_get = "GetOption",
						child_set = "SetOption",
						order = 50,
						args = {
							left = {
								type = "number",
								name = L["Left"],
								desc = L["Stretches or shrinks the background texture."],
								step = 1,
								bigStep = 5,
								min = -50,
								max = 50,
								passValue4 = "left",
								stepBasis = 5,
								order = 10,
							},
							right = {
								type = "number",
								name = L["Right"],
								desc = L["Stretches or shrinks the background texture."],
								step = 1,
								bigStep = 5,
								min = -50,
								max = 50,
								passValue4 = "right",
								stepBasis = 5,
								order = 20,
							},
							top = {
								type = "number",
								name = L["Top"],
								desc = L["Stretches or shrinks the background texture."],
								step = 1,
								bigStep = 5,
								min = -50,
								max = 50,
								passValue4 = "top",
								stepBasis = 5,
								order = 30,
							},
							bottom = {
								type = "number",
								name = L["Bottom"],
								desc = L["Stretches or shrinks the background texture."],
								step = 1,
								bigStep = 5,
								min = -50,
								max = 50,
								passValue4 = "bottom",
								stepBasis = 5,
								order = 40,
							},
						},
					},
				},
			},
		},
	},
	BarElements = {
		type = "group",
		desc = L["Configures bar element options, such as the icon, timer, label, and so on."],
		name = L["Bar elements"],
		order = 70,
		args = {
			Status = {
				type = "group",
				desc = L["Configures the statusbar options."],
				name = L["Statusbar"],
				order = 10,
				groupType = "inline",
				child_get = "GetOption",
				child_set = "SetOption",
				child_passValue = "Bar",
				child_passValue2 = "Status",
				args = {
					Texture = {
						type = "choice",
						name = L["Texture"],
						desc = L["Sets the statusbar texture."],
						choices = MediaHandler:List("statusbar"),
						passValue3 = "Texture",
						order = 10,
					},
					LeftSpacing = {
						type = "number",
						name = L["Left Spacing"],
						desc = L["Sets the spacing between the statusbar texture and the edge of the bar."],
						min = -5,
						max = 5,
						step = 1,
						bigStep = 1,
						passValue3 = "LeftSpacing",
						order = 20,
					},
					RightSpacing = {
						type = "number",
						name = L["Right Spacing"],
						desc = L["Sets the spacing between the statusbar texture and the edge of the bar."],
						min = -5,
						max = 5,
						step = 1,
						bigStep = 1,
						passValue3 = "RightSpacing",
						order = 30,
					},
					TopSpacing = {
						type = "number",
						name = L["Top Spacing"],
						desc = L["Sets the spacing between the statusbar texture and the edge of the bar."],
						min = -5,
						max = 5,
						step = 1,
						bigStep = 1,
						passValue3 = "TopSpacing",
						order = 40,
					},
					BottomSpacing = {
						type = "number",
						name = L["Bottom Spacing"],
						desc = L["Sets the spacing between the statusbar texture and the edge of the bar."],
						min = -5,
						max = 5,
						step = 1,
						bigStep = 1,
						passValue3 = "BottomSpacing",
						order = 50,
					},
				},
			},
			Icon = {
				type = "group",
				desc = L["Configures icon options."],
				name = L["Icon"],
				order = 30,
				groupType = "inline",
				child_get = "GetOption",
				child_set = "SetOption",
				child_passValue = "Bar",
				child_passValue2 = "Icon",
				args = {
					Visible = {
						type = "boolean",
						name = L["Visible"],
						desc = L["Sets the visibilty of the bar icon."],
						passValue3 = "Visible",
						order = 10,
					},
					Location = {
						type = "choice",
						name = L["Location"],
						desc = L["Sets the anchor location of the bar icon."],
						choices = {
							LEFT = L["Left"],
							RIGHT = L["Right"],
						},
						passValue3 = "Location",
						order = 20,
					},
				},
			},
			Label = {
				type = "group",
				desc = L["Configures icon options."],
				name = L["Label"],
				order = 40,
				groupType = "inline",
				child_get = "GetOption",
				child_set = "SetOption",
				child_passValue = "Bar",
				child_passValue2 = "Label",
				args = {
					Font = {
						type = "choice",
						name = L["Font"],
						desc = L["Sets the label font."],
						choices = MediaHandler:List("font"),
						passValue3 = "Font",
						order = 10,
					},
					FontSize = {
						type = "number",
						name = L["Font Size"],
						desc = L["Sets the size of the label font."],
						min = 9,
						max = 24,
						step = 1,
						bigStep = 1,
						passValue3 = "FontSize",
						order = 20,
					},
					JustifyH = {
						type = "choice",
						name = L["Horizontal justify"],
						desc = L["Justifies the label position."],
						choices = {
							LEFT = L["Left"],
							CENTER = L["Center"],
							RIGHT = L["Right"],
						},
						passValue3 = "JustifyH",
						order = 30,
					},
					JustifyV = {
						type = "choice",
						name = L["Vertical justify"],
						desc = L["Justifies the label position."],
						choices = {
							TOP = L["Top"],
							CENTER = L["Center"],
							BOTTOM = L["Bottom"],
						},
						passValue3 = "JustifyV",
						order = 40,
					},
					Adjust = {
						type = "number",
						name = L["Horizontal adjust"],
						desc = L["Slightly adjusts the horizontal positioning of the label text."],
						min = -50,
						max = 50,
						step = 1,
						bigStep = 5,
						passValue3 = "Adjust",
						order = 50,
					},
					Color = {
						type = "color",
						name = L["Color"],
						desc = L["Sets the coloring used on the label text."],
						hasAlpha = true,
						passValue = "colorTable",
						passValue2 = "Bar",
						passValue3 = "Label",
						passValue4 = "Color",
						order = 60,
					},
				},
			},
			Timer = {
				type = "group",
				desc = L["Configures icon options."],
				name = L["Timer"],
				order = 50,
				groupType = "inline",
				child_get = "GetOption",
				child_set = "SetOption",
				child_passValue = "Bar",
				child_passValue2 = "Timer",
				args = {
					Font = {
						type = "choice",
						name = L["Font"],
						desc = L["Sets the timer font."],
						choices = MediaHandler:List("font"),
						passValue3 = "Font",
						order = 10,
					},
					FontSize = {
						type = "number",
						name = L["Font Size"],
						desc = L["Sets the size of the timer font."],
						min = 9,
						max = 24,
						step = 1,
						bigStep = 1,
						passValue3 = "FontSize",
						order = 20,
					},
					JustifyH = {
						type = "choice",
						name = L["Horizontal justify"],
						desc = L["Justifies the timer position."],
						choices = {
							LEFT = L["Left"],
							CENTER = L["Center"],
							RIGHT = L["Right"],
						},
						passValue3 = "JustifyH",
						order = 30,
					},
					JustifyV = {
						type = "choice",
						name = L["Vertical justify"],
						desc = L["Justifies the timer position."],
						choices = {
							TOP = L["Top"],
							CENTER = L["Center"],
							BOTTOM = L["Bottom"],
						},
						passValue3 = "JustifyV",
						order = 40,
					},
					Adjust = {
						type = "number",
						name = L["Horizontal adjust"],
						desc = L["Slightly adjusts the horizontal positioning of the timer text."],
						min = -50,
						max = 50,
						step = 1,
						bigStep = 5,
						passValue3 = "Adjust",
						order = 50,
					},
					Color = {
						type = "color",
						name = L["Color"],
						desc = L["Sets the coloring used on the timer text."],
						hasAlpha = true,
						passValue = "colorTable",
						passValue2 = "Bar",
						passValue3 = "Timer",
						passValue4 = "Color",
						order = 60,
					},
				},
			},
			Spark = {
				type = "group",
				desc = L["Configures spark options."],
				name = L["Spark"],
				order = 60,
				groupType = "inline",
				child_get = "GetOption",
				child_set = "SetOption",
				child_passValue = "Bar",
				child_passValue2 = "Spark",
				args = {
					Visible = {
						type = "boolean",
						name = L["Visible"],
						desc = L["Sets the visibilty of the bar spark."],
						passValue3 = "Visible",
						order = 10,
					},
					Texture = {
						type = "choice",
						name = L["Texture"],
						desc = L["Sets the spark texture for bars to use."],
						choices = MediaHandler:List("spark"),
						passValue3 = "Texture",
						order = 20,
					},
					Blend = {
						type = "choice",
						name = L["Blend"],
						desc = L["Sets the blending used to render the spark texture."],
						choices = blendModes,
						passValue3 = "Blend",
						order = 30,
					},
					Height = {
						type = "number",
						name = L["Height"],
						desc = L["Sets the height scale used on the spark texture."],
						isPercent = true,
						min = .1,
						max = 3,
						step = .01,
						bigStep = .1,
						passValue3 = "Height",
						order = 40,
					},
					Width = {
						type = "number",
						name = L["Width"],
						desc = L["Sets the width scale used on the spark texture."],
						isPercent = true,
						min = .1,
						max = 3,
						step = .01,
						bigStep = .1,
						passValue3 = "Width",
						order = 50,
					},
					Color = {
						type = "color",
						name = L["Color"],
						desc = L["Sets the coloring used on the spark texture."],
						hasAlpha = true,
						passValue = "colorTable",
						passValue2 = "Bar",
						passValue3 = "Spark",
						passValue4 = "Color",
						order = 60,
					},
				},
			},
		},
	},
	FilterOptions = {
		type = "group",
		desc = L["Controls which bars should be rejected from this frame."],
		name = L["Filtering options"],
		order = 90,
		args = {
			Categories = {
				type = "group",
				desc = L["Rejects bars that aren't part of an allowed category."],
				name = L["Categories"],
				order = 10,
				groupType = "inline",
				child_get = "GetOption",
				child_set = "SetOption",
				args = {
					Categories = {
						type = "boolean",
						name = L["Filter by category"],
						desc = L["Whether or not this frame should filter bars that don't meet the selected categories."],
						passValue = "Categories",
						order = 10,
					},
					BossAbilities = {
						type = "multichoice",
						name = L["Boss abilities"],
						desc = L["A checkmark signifies that the category should be rejected, otherwise the category is allowed."],
						choices = {
							Cast = L["Cast"],
							Cooldown = L["Cooldown"],
							Buff = L["Buff"],
							Debuff = L["Debuff"],
							Timeless = L["Timeless"],
							Count = L["Count"],
						},
						choiceDescs = {
							Cast = L["Casting progress of a boss' ability"],
							Cooldown = L["Boss' cooldown remaining"],
							Buff = L["Duration of a boss' buff"],
							Debuff = L["Duration of a boss' buff"],
							Timeless = L["Bars without time"],
							Count = L["Bars that display a count"],
						},
						passValue = "CategoriesList",
						passValue2 = "Boss Abilities",
						order = 20,
					},
					ModuleCategories = {
						type = "group",
						desc = L["Displays each plugin's categories list and allows them to be rejected."],
						name = L["Plugin categories"],
						order = 30,
						groupType = "inline",
						hidden = function() return not next(RefreshConfigModuleCategories()) end, --serves as both a proxy and to hide these options when there's no plugins
						child_get = "GetOption",
						child_set = "SetOption",
						child_passValue = "CategoriesList",
						child_passValue2 = "Modules",
						child_desc = L["A checkmark signifies that the category should be rejected, otherwise the category is allowed."],
						args = moduleCategories,
					}, 
				},
			},
			Labels = {
				type = "group",
				desc = L["Rejects bars that exactly match a certain label."],
				name = L["Labels"],
				order = 20,
				groupType = "inline",
				args = {
					Labels = {
						type = "boolean",
						name = L["Filter by label"],
						desc = L["Whether or not this frame should reject bars that exactly match a certain label."],
						get = "GetOption",
						set = "SetOption",
						passValue = "Labels",
						order = 10,
					},
					NewLabel = {
						type = "string",
						name = L["New label"],
						desc = L["Adds a new label to be rejected."],
						get = false,
						set = "AddFilteredLabel",
						order = 20,
						validate = function(name)
							if name == "" then
								return false, L["The label name cannot be empty."]
							end
							return true
						end,
						usage = L["Enter the EXACT name of the label you wish to reject and press enter."],
					},
					LabelsList = {
						type = "multichoice",
						name = L["Currently filtered"],
						desc = L["Labels appearing in this list would be rejected. Clicking on a label will remove it from the list."],
						choices = "GetFilteredLabels",
						get = ReturnTrue,
						set = "RemoveFilteredLabel",
						order = 30,
					},
				},
			},
			ByMaxLength = {
				type = "group",
				desc = L["Rejects bars that are over a certain length."],
				name = L["Max Length"],
				order = 30,
				groupType = "inline",
				child_get = "GetOption",
				child_set = "SetOption",
				args = {
					ByMaxLength = {
						type = "boolean",
						name = L["Filter By Max Length"],
						desc = L["Whether or not this frame should reject bars over a certain amount of time."],
						passValue = "ByMaxLength",
						order = 10,
					},
					MaxLength = {
						type = "number",
						name = L["Max Length"],
						desc = L["The maximum amount of time, in seconds, before a bar will be rejected."],
						step = 1,
						bigStep = 10,
						stepBasis = 10,
						min = 0,
						max = 600,
						passValue = "MaxLength",
						order = 20,
					},
				},
			},
			ByMinLength = {
				type = "group",
				desc = L["Rejects bars that are under a certain length."],
				name = L["Min Length"],
				order = 40,
				groupType = "inline",
				child_get = "GetOption",
				child_set = "SetOption",
				args = {
					ByMinLength = {
						type = "boolean",
						name = L["Filter By Min Length"],
						desc = L["Whether or not this frame should reject bars under a certain amount of time."],
						passValue = "ByMinLength",
						order = 10,
					},
					MinLength = {
						type = "number",
						name = L["Min Length"],
						desc = L["The minimum amount of time, in seconds, before a bar will be rejected."],
						step = 1,
						bigStep = 10,
						stepBasis = 10,
						min = 0,
						max = 600,
						passValue = "MinLength",
						order = 20,
					},
				},
			},
			Custom = {
				type = "group",
				desc = L["Rejects bars that partially match a word."],
				name = L["Custom filters"],
				order = 50,
				groupType = "inline",
				args = {
					Labels = {
						type = "boolean",
						name = L["Custom filtering"],
						desc = L["Whether or not this frame should reject bars that partially match a word."],
						get = "GetOption",
						set = "SetOption",
						passValue = "Custom",
						order = 10,
					},
					NewLabel = {
						type = "string",
						name = L["New custom filter"],
						desc = L["Adds a filter, that if matched will reject the bar. Will also match regular expressions. For example entering 'Innervate' would reject every bar that has 'Innervate' in it."],
						get = false,
						set = "AddCustomFilter",
						order = 20,
						validate = function(name)
							if name == "" then
								return false, L["The custom filter cannot be empty."]
							end
							return true
						end,
						usage = L["Enter the custom filter you wish to add and press enter."],
					},
					CustomList = {
						type = "multichoice",
						name = L["Current filters"],
						desc = L["Bars matching these filters will be rejected. Clicking on a filter will remove it from the list."],
						choices = "GetCustomFilters",
						get = ReturnTrue,
						set = "RemoveCustomFilter",
						order = 30,
					},
				},
			},
			CustomException = {
				type = "group",
				desc = L["Any bar that doesn't match one of these filters will be rejected."],
				name = L["Custom exception filters"],
				order = 60,
				groupType = "inline",
				args = {
					Labels = {
						type = "boolean",
						name = L["Exception filtering"],
						desc = L["Whether or not this frame should reject bars that don't match a word."],
						get = "GetOption",
						set = "SetOption",
						passValue = "CustomException",
						order = 10,
					},
					NewLabel = {
						type = "string",
						name = L["New custom filter"],
						desc = L["Adds a filter, that unless a bar matches it the bar will be rejected. Will also match regular expressions. For example, if 'Resurrection' was entered then only bars that have Resurrection in their label would be allowed."],
						get = false,
						set = "AddCustomException",
						order = 20,
						validate = function(name)
							if name == "" then
								return false, L["The custom filter cannot be empty."]
							end
							return true
						end,
						usage = L["Enter the custom filter you wish to add and press enter."],
					},
					CustomExceptionList = {
						type = "multichoice",
						name = L["Current filters"],
						desc = L["Bars not matching these filters will be rejected. Clicking on a filter will remove it from the list."],
						choices = "GetCustomExceptions",
						get = ReturnTrue,
						set = "RemoveCustomException",
						order = 30,
					},
				},
			},
		},
	},
	SoundOptions = {
		type = "group",
		desc = L["Let's you choose auditory alerts when bars reach a certain time."],
		name = L["Sound options"],
		order = 95,
		child_set = "SetSoundOption",
		child_get = "GetSoundOption",
		args = {
			DisableAllSounds = {
				type = "execute",
				name = L["Disable all sounds"],
				desc = L["Disables sounds for all categories and subcategories."],
				func = "SetAllSounds",
				buttonText = L["Disable all sounds"],
				passValue = false,
				order = 2,
			},
			EnableAllSounds = {
				type = "execute",
				name = L["Enable all sounds"],
				desc = L["Enables sounds for all categories and subcategories."],
				func = "SetAllSounds",
				buttonText = L["Enable all sounds"],
				passValue = true,
				order = 3,
			},
			CurrentCategory = {
				type = "choice",
				name = L["Select category"],
				desc = L["Selects the category to configure sound options for."],
				set = "SetCurrentSoundCategory",
				get = "GetCurrentSoundCategory",
				choices = function()
					local dict = newDict()
					dict["Boss Abilities"]--#NO_LOCALIZE
						= L["Boss Abilities"]
					for name, module in Core:IterateModules("Plugin") do
						if module:GetCategories() then
							dict[name] = name
						end
					end
					return "@dict", unpackDictAndDel(dict)
				end,
				order = 8,
			},
			CurrentSubCategory = {
				type = "choice",
				name = L["Select subcategory"],
				desc = L["Selects the subcategory to configure sound options for."],
				set = "SetCurrentSoundSubCategory",
				get = "GetCurrentSoundSubCategory",
				choices = "GetSoundSubCategoryChoices",
				order = 9,
			},
			SoundByPercent = {
				type = "boolean",
				name = L["Sound alert by percent"],
				desc = L["Enables an alert when bars are under a certain percentage in time. For example, if this is set to 50% and a 2 minute bar goes under 1 minute left a sound will be played."],
				order = 10,
				passValue = "SoundByPercent",
			},
			SoundPercent = {
				type = "number",
				name = L["Alert percentage"],
				desc = L["If the bar time goes under this percentage of time a sound will be played."],
				isPercent = true,
				step = .01,
				bigStep = .05,
				stepBasis = .05,
				passValue = "SoundPercent",
				order = 20,
			},
			SoundByTime = {
				type = "boolean",
				name = L["Sound alert by absolute time"],
				desc = L["Enables an alert when bars are under a certain amount of seconds. If sound alert by percent is also enabled whichever criteria is met first will trigger the sound."],
				order = 30,
				passValue = "SoundByTime",
			},
			SoundTime = {
				type = "number",
				name = L["Alert time"],
				desc = L["If the bar time goes under this time in seconds a sound will be played."],
				step = 1,
				bigStep = 5,
				stepBasis = 5,
				min = 0,
				max = 600,
				passValue = "SoundTime",
				order = 40,
			},
			SoundToPlay = {
				type = "choice",
				name = L["Sound to play"],
				desc = L["Sets the sound that will be played when the criteria above is met."],
				choices = MediaHandler:List("sound"),
				passValue = "SoundToPlay",
				order = 50,
			},
			TestSound = {
				type = "execute",
				name = L["Test sound"],
				desc = L["Plays the currently selected sound."],
				func = "PlayCurrentSound",
				buttonText = L["Test Sound"],
				order = 60,
			},
			Labels = {
				type = "group",
				desc = L["Sets sounds for certain labels."],
				name = L["Labels"],
				order = 70,
				groupType = "inline",
				args = {
					NewLabel = {
						type = "string",
						name = L["New label"],
						desc = L["Adds a new label to be have a custom sound, instead of a category sound."],
						get = false,
						set = "AddCustomSoundLabel",
						order = 20,
						validate = function(name)
							if name == "" then
								return false, L["The label name cannot be empty."]
							end
							return true
						end,
						usage = L["Enter the EXACT name of the label you wish to customize a sound for and press enter."],
					},
					LabelsList = {
						type = "choice",
						name = L["Current labels"],
						desc = L["Labels that currently use a special sound instead of a category sound. Select to modify."],
						choices = "GetCustomSoundLabels",
						get = "GetCurrentCustomSoundLabel",
						set = "SetCurrentCustomSoundLabel",
						order = 30,
					},
					SoundToPlay = {
						type = "choice",
						name = L["Sound to play"],
						desc = L["Sets the sound that will be played for the label selected above."],
						set = "SetCurrentCustomSound",
						get = "GetCurrentCustomSound",
						disabled = "~GetCurrentCustomSoundLabel",
						choices = MediaHandler:List("sound"),
						order = 50,
					},
					TestSound = {
						type = "execute",
						name = L["Test sound"],
						desc = L["Plays the currently selected sound."],
						func = "PlayCurrentCustomSound",
						buttonText = L["Test Sound"],
						disabled = "~GetCurrentCustomSoundLabel",
						order = 60,
					},
					RemoveLabel = {
						type = "execute",
						name = L["Remove label"],
						desc = L["Removes the currently selected label, returning it to it's default category based sound options."],
						func = "RemoveCustomSoundLabel",
						buttonText = L["Remove custom label"],
						disabled = "~GetCurrentCustomSoundLabel",
						order = 70,
					},

				},
			},
		},
	},
	Miscellaneous = {
		type = "group",
		desc = L["Miscellaneous options, such as label recoloring, and theme options."],
		name = L["Misc options"],
		order = 100,
		args = {
			ColoredLabels = {
				type = "group",
				desc = L["Allows certain labels to be colored, so that they stand out."],
				name = L["Colored labels"],
				order = 10,
				groupType = "inline",
				args = {
					NewLabel = {
						type = "string",
						name = L["New label"],
						desc = L["Adds a new label to color."],
						get = false,
						set = "AddColoredLabel",
						order = 10,
						validate = function(name)
							if name == "" then
								return false, L["The label name cannot be empty."]
							end
							return true
						end,
						usage = L["Enter the EXACT name of the label you wish to color and press enter."],
					},
					EditLabel = {
						type = "choice",
						name = L["Current labels"],
						desc = L["Select one of the labels from the dropdown to edit it."],
						get = "GetCurrentColoredLabel",
						set = "SetCurrentColoredLabel",
						order = 20,
						choices = "GetColoredLabelList",
					},
					Color = {
						type = "color",
						name = L["Label color"],
						desc = L["Sets the color of bars that match this label exactly."],
						disabled = "~GetCurrentColoredLabel",
						set = "AddColoredLabel",
						get = "GetColoredLabel",
						passValue = false,
						order = 30,
					},
					Reset = {
						type = "execute",
						name = L["Remove label"],
						desc = L["Removes the currently editted label, resetting the bar coloring back to default."],
						func = "RemoveColoredLabel",
						buttonText = L["Remove label"],
						disabled = "~GetCurrentColoredLabel",
						order = 40,
					},
				},
			},
			ThemeOptions = {
				type = "group",
				desc = L["Themes allow you to save all the settings from a frame and apply them to other frames."],
				name = L["Theme options"],
				groupType = "inline",
				order = 20,
				args = {
					SaveTheme = {
						type = "group",
						desc = L["Sets up options that allow you to save the current frame settings as a theme so they can be easily re-applied to this or another frame later."],
						name = L["Save theme options"],
						groupType = "inline",
						order = 10,
						args = {
							SaveFilters = {
								type = "boolean",
								name = L["Save filter settings"],
								desc = L["Saves filter settings from this frame into the theme.\n\nThis option is not saved between sessions."],
								order = 10,
								passValue = "Filters",
								set = SetThemeOption,
								get = GetThemeOption,
							},
							SaveElements = {
								type = "boolean",
								name = L["Save bar element settings"],
								desc = L["Saves bar element settings (such as icon, font, and so on) from this frame into the theme.\n\nThis option is not saved between sessions."],
								order = 20,
								passValue = "Elements",
								set = SetThemeOption,
								get = GetThemeOption,
							},
							SaveTransfer = {
								type = "boolean",
								name = L["Save transfer options"],
								desc = L["Saves frame transfer settings from this frame into the theme.\n\nThis option is not saved between sessions."],
								order = 30,
								passValue = "Transfer",
								set = SetThemeOption,
								get = GetThemeOption,
							},
							SaveSound = {
								type = "boolean",
								name = L["Save sound options"],
								desc = L["Saves sound options from this frame into the theme.\n\nThis option is not saved between sessions."],
								order = 40,
								passValue = "Sound",
								set = SetThemeOption,
								get = GetThemeOption,
							},
							SaveTheme = {
								type = "string",
								name = L["Save current as:"],
								desc = L["Saves the current settings from this frame as a theme using the options set above."],
								get = false,
								set = "SaveTheme",
								order = 50,
								passValue = themeOpts,
								usage = L["Enter the name of the theme you wish to save the frame as and press enter."],
							},
						},
					},				
					ApplyTheme = {
						type = "choice",
						name = L["Apply theme"],
						desc = L["Applies all the settings saved in a theme to this frame. WARNING: CURRENT SETTINGS WILL BE LOST."],
						get = Noop,
						set = "ApplyTheme",
						choices = function()
							local list = newList()
							for name in pairs(ThemeList) do
								list[#list + 1] = name
							end
							return "@list", unpackListAndDel(list)
						end,
						order = 20,
					},
					KeepAnchorPoints = {
						type = "boolean",
						name = L["Keep anchor points"],
						desc = L["Keeps the current frame's anchor points when applying a theme."],
						order = 30,
						passValue = "KeepAnchorPoints",
						set = "SetOption",
						get = "GetOption",
					},
					MergeWithExisting = {
						type = "boolean",
						name = L["Merge theme with existing settings"],
						desc = L["Merges the current frame settings with the applying theme, otherwise the frame will be reset to defaults before the theme is applied."],
						order = 40,
						passValue = "MergeWithExisting",
						set = "SetOption",
						get = "GetOption",
					},
					SaveFilters = {
						type = "boolean",
						name = L["Save filters"],
						desc = L["When applying a theme the current frame filters will be saved, instead using filters saved in the theme."],
						order = 50,
						passValue = "SaveFilters",
						set = "SetOption",
						get = "GetOption",
					},
				},
			},
		},
	},
}